import os
import h5py
import numpy as np

class DatasetWriter:
    def __init__(self, dataset_name, num_rows, dtypes, hdf_file_path, buffer_size,
                 force=False):
        if not force and os.path.exists(hdf_file_path):
            raise Exception(
                'DatasetWriter::__init__: {} already exists.'.format(
                    hdf_file_path))
        self.hdf_file_path = hdf_file_path
        self.db = h5py.File(hdf_file_path, 'w')
        self.dtypes = dtypes
        self.dataset = self.db.create_dataset(dataset_name, (num_rows,), dtype=dtypes)

        if buffer_size <= 0:
            raise Exception(
                'DatasetWriter::__init__: Buffer size must be > 0.')
        self.buffer_size = buffer_size
        self.buffer = np.recarray((buffer_size,), dtype=dtypes)
        self.buffer_idx = 0
        self.dataset_idx = 0

    def add(self, row):
        if self.dataset_idx == len(self.dataset):
            print(f'Dataset full. Not adding request row: {row}.')
            return

        # Check if the buffer is full and, if so, flush the buffer.
        if self.buffer_idx == self.buffer.shape[0]:
            self.flush()
        row_tuple = tuple([row[name] for name in list(self.buffer.dtype.names)])
        self.buffer[self.buffer_idx] = row_tuple
        self.buffer_idx += 1


    def clear_buffer(self):
        self.buffer_idx = 0
        self.buffer = np.recarray((self.buffer_size,), dtype=self.dtypes)

    def flush(self):
        if self.buffer_idx != 0:
            self.dataset[self.dataset_idx:self.dataset_idx+self.buffer_idx] = self.buffer[0:self.buffer_idx]
            self.dataset_idx += self.buffer_idx
            self.clear_buffer()

    def close(self):
        self.flush()
        self.db.close()



# Taken from Deep Learning for Computer Vision with Python, Starter Bundle

from tensorflow.keras.preprocessing.image import img_to_array


class ImageToArrayPreprocessor:
    def __init__(self, data_format=None):
        self.data_format = data_format

    def preprocess(self, imgs):
        return [
            img_to_array(img, data_format=self.data_format) for img in imgs
        ]



import cv2
import os
import numpy as np


class FaceExtractor:
    def __init__(self, prototxt_file_path, model_file_path):
        if not os.path.exists(prototxt_file_path):
            raise Exception(
                'FaceExtractor::__init__: Prototxt file {} does not exist.'.
                format(prototxt_file_path))

        if not os.path.exists(model_file_path):
            raise Exception(
                'FaceExtractor::__init__: Model file {} does not exist.'.
                format(model_file_path))

        self.face_detection_model = cv2.dnn.readNetFromCaffe(
            prototxt_file_path, model_file_path)

    def extract_from_img(self, img):
        (h, w) = img.shape[:2]
        blob = cv2.dnn.blobFromImage(cv2.resize(img, (300, 300)), 1.0,
                                     (300, 300))
        self.face_detection_model.setInput(blob)
        detections = self.face_detection_model.forward()
        confidence = detections[0, 0, 0, 2]
        box = detections[0, 0, 0, 3:7] * np.array([w, h, w, h])
        (startX, startY, endX, endY) = box.astype('int')
        face = img[startY:endY, startX:endX]

        return (confidence, face)



import face_recognition
import numpy as np
import cv2
import matplotlib.pyplot as plt


class FacePoseExtractor:
    def __init__(self):
        # Based on http://aifi.isr.uc.pt/Downloads/OpenGL/glAnthropometric3DModel.cpp.
        model_chin_pts = [(-7.308957, 0.913869, -6.76343),
                          (-6.775290, -0.730814, -6.776229),
                          (-5.665918, -3.286078, -5.740479),
                          (-5.011779, -4.876396, -5.715469),
                          (-4.056931, -5.947019, -5.1272009999999995),
                          (-1.833492, -7.056977, -2.7021549999999994),
                          (0.000000, -7.415691, -2.692996),
                          (1.833492, -7.056977, -2.7021549999999994),
                          (4.056931, -5.947019, -5.1272009999999995),
                          (5.011779, -4.876396, -5.715469),
                          (5.665918, -3.286078, -5.740479),
                          (6.775290, -0.730814, -6.776229),
                          (7.308957, 0.913869, -6.76343)]
        model_left_eyebrow_pts = [
            (-6.825897, 6.760612, -2.361287999999999),
            (-6.137002, 7.271266, -1.5626069999999999),
            (-4.861131, 7.878672, -0.16215499999999938),
            (-2.533424, 7.878085, 0.6876040000000003),
            (-1.330353, 7.122144, 0.1403150000000002),
        ]
        model_right_eyebrow_pts = [(1.330353, 7.122144, 0.1403150000000002),
                                   (2.533424, 7.878085, 0.6876040000000003),
                                   (4.861131, 7.878672, -0.16215499999999938),
                                   (6.137002, 7.271266, -1.5626069999999999),
                                   (6.825897, 6.760612, -2.361287999999999)]
        model_nose_bridge_pts = [(0, 5.662829, 0.8906200000000002),
                                 (0, 4.347349, -0.0980836699999994),
                                 (0, 3.031869, -1.08678733),
                                 (0, 1.716389, -2.0754909999999995)]
        model_nose_tip_pts = [(-1.930245, 0.424351, -0.8490539999999998),
                              (-0.746313, 0.348381, -0.500203),
                              (0.000000, 0.000000, 0.0),
                              (0.746313, 0.348381, -0.500203),
                              (1.930245, 0.424351, -0.8490539999999998)]
        model_left_lip_corner_pt = (-2.774015, -2.080775, -1.714899)
        model_right_lip_corner_pt = (2.774015, -2.080775, -1.714899)

        model_inner_face_pts = []
        model_inner_face_pts.extend(model_left_eyebrow_pts)
        model_inner_face_pts.extend(model_right_eyebrow_pts)
        model_inner_face_pts.extend(model_nose_bridge_pts)
        model_inner_face_pts.extend(model_nose_tip_pts)
        model_inner_face_pts.append(model_left_lip_corner_pt)
        model_inner_face_pts.append(model_right_lip_corner_pt)
        model_outer_face_pts = model_chin_pts
        self.model_outer_face_pts = np.array(model_outer_face_pts,
                                             dtype='double')
        self.model_inner_face_pts = np.array(model_inner_face_pts,
                                             dtype='double')
        self.model_whole_face_pts = np.array(model_outer_face_pts +
                                             model_inner_face_pts,
                                             dtype='double')
        # plt.plot([pt[0] for pt in self.model_whole_face_pts],
        #          [pt[1] for pt in self.model_whole_face_pts],
        #          'ro',
        #          alpha=0.5)
        # plt.show()
        # self.model_whole_face_pts = np.array(
        #     model_inner_face_pts, dtype='double')  # TODO: remove, testing

    def get_face_poses(self, img):
        camera_matrix = self.get_camera_matrix(img)
        face_landmarks_list = face_recognition.face_landmarks(img)
        whole_face_pts_list = []
        inner_face_pts_list = []
        nose_pts_list = []
        for face_landmarks in face_landmarks_list:
            outer_face_pts = face_landmarks['chin'][2:15]
            inner_face_pts = []
            inner_face_pts.extend(face_landmarks['left_eyebrow'])
            inner_face_pts.extend(face_landmarks['right_eyebrow'])
            inner_face_pts.extend(face_landmarks['nose_bridge'])
            inner_face_pts.extend(face_landmarks['nose_tip'])
            inner_face_pts.append(face_landmarks['top_lip'][0])
            inner_face_pts.append(face_landmarks['top_lip'][7])
            whole_face_pts_list.append(
                np.array(outer_face_pts + inner_face_pts, dtype='double'))
            # whole_face_pts_list.append(np.array(
            #     inner_face_pts, dtype='double'))  # TODO: remove, testing
            inner_face_pts_list.append(np.array(inner_face_pts,
                                                dtype='double'))
            nose_pts_list.append(face_landmarks['nose_tip'][2])

        poses = []
        for whole_face_pts, inner_face_pts, nose_pt in zip(
                whole_face_pts_list, inner_face_pts_list, nose_pts_list):
            (_, whole_face_rotation_vector,
             whole_face_translation_vector) = cv2.solvePnP(
                 self.model_whole_face_pts,
                 whole_face_pts,
                 camera_matrix,
                 np.zeros((4, 1)),
                 flags=cv2.SOLVEPNP_ITERATIVE)
            (_, inner_face_rotation_vector,
             inner_face_translation_vector) = cv2.solvePnP(
                 self.model_inner_face_pts,
                 inner_face_pts,
                 camera_matrix,
                 np.zeros((4, 1)),
                 flags=cv2.SOLVEPNP_ITERATIVE)

            poses.append({
                'whole': {
                    'rotation': whole_face_rotation_vector,
                    'translation': whole_face_translation_vector
                },
                'inner': {
                    'rotation': inner_face_rotation_vector,
                    'translation': inner_face_translation_vector
                },
                'nose_point': nose_pt
            })

            for p in whole_face_pts_list[0]:
                cv2.circle(img, (int(p[0]), int(p[1])), 3, (0, 0, 255), -1)

        return poses

    def get_camera_matrix(self, img):
        height = img.shape[0]
        width = img.shape[1]
        focal_length = width
        center = (width / 2, height / 2)
        camera_matrix = np.array([[focal_length, 0, center[0]],
                                  [0, focal_length, center[1]], [0, 0, 1]],
                                 dtype='double')

        return camera_matrix

    def draw_pose_lines(self, img):
        poses = self.get_face_poses(img)
        camera_matrix = self.get_camera_matrix(img)
        for pose in poses:
            (whole_face_nose_end_point2D,
             jacobian) = cv2.projectPoints(np.array([
                 (0.0, 0.0, 1000.0)
             ]), pose['whole']['rotation'], pose['whole']['translation'],
                                           camera_matrix, np.zeros((4, 1)))
            (inner_face_nose_end_point2D,
             jacobian) = cv2.projectPoints(np.array([
                 (0.0, 0.0, 1000.0)
             ]), pose['inner']['rotation'], pose['inner']['translation'],
                                           camera_matrix, np.zeros((4, 1)))

            print(pose)
            print('-------')

            p1 = (int(pose['nose_point'][0]), int(pose['nose_point'][1]))
            whole_face_p2 = (int(whole_face_nose_end_point2D[0][0][0]),
                             int(whole_face_nose_end_point2D[0][0][1]))
            inner_face_p2 = (int(inner_face_nose_end_point2D[0][0][0]),
                             int(inner_face_nose_end_point2D[0][0][1]))
            print(f'inner_face_p2: {inner_face_p2}.')

            cv2.line(img, p1, whole_face_p2, (255, 0, 0), 2)
            cv2.line(img, p1, inner_face_p2, (0, 0, 255), 2)

        return img



import PIL


class CropPreprocessor:
    def __init__(self, box_origin, box_width, box_height):
        self.box_origin = box_origin
        self.box_width = box_width
        self.box_height = box_height

    def preprocess(self, imgs):
        crops = []

        for img in imgs:
            if isinstance(img, PIL.Image.Image):
                crop = img.crop(box=(self.box_origin[0], self.box_origin[1],
                                     self.box_origin[0] + self.box_width,
                                     self.box_origin[1] + self.box_height))
            else:
                crop = img[self.box_origin[1]:self.box_origin[1] +
                           self.box_height, self.
                           box_origin[0]:self.box_origin[0] + self.box_width]
            crops.append(crop)

        return crops



# import the necessary packages
import cv2


# Taken from the pyimagesearch Practitioner Bundle.
class MeanPreprocessor:
    def __init__(self, r_mean, g_mean, b_mean):
        # store the Red, Green, and Blue channel averages across a
        # training set
        self.r_mean = r_mean
        self.g_mean = g_mean
        self.b_mean = b_mean

    def preprocess(self, seq):
        seq_length = seq.shape[0]
        for i in range(seq_length):
            # split the image into its respective Red, Green, and Blue
            # channels
            (b, g, r) = cv2.split(seq[i].astype("float32"))

            # subtract the means for each channel
            r -= self.r_mean
            g -= self.g_mean
            b -= self.b_mean

            # merge the channels back together
            seq[i] = cv2.merge([b, g, r])

        return seq



import unittest
from datasetio.datasetwriter import DatasetWriter
from datasetio.datasetgenerator import DatasetGenerator
import h5py
import os
import numpy as np
import string
import random
import math


class TestDatasetGenerator(unittest.TestCase):
    def setUp(self):
        self.feat_length = 5
        self.seq_length = 5
        self.buffer_size = 5
        self.num_rows = 50
        self.dataset_file_path = 'test.hdf'
        self.dataset_name = 'test'

        self.dtypes = [('feat_seq', 'float', (self.seq_length,
                                              self.feat_length)),
                       ('label', 'int'), ('file', h5py.string_dtype())]
        self.dataset_writer = DatasetWriter('test', self.num_rows, self.dtypes,
                                            self.dataset_file_path,
                                            self.buffer_size)
        self.taken_files = set()

    def tearDown(self):
        os.remove(self.dataset_file_path)

    def initialize_expected_rows(self):
        expected_rows = []
        for i in range(0, self.num_rows):
            zero_features = np.zeros((self.seq_length, self.feat_length))
            row = self.generate_row(zero_features, 0, '')
            expected_rows.append(row)

        return expected_rows

    def generate_row(self, features, label, file):
        return {'feat_seq': features, 'label': label, 'file': file}

    def generate_random_row(self):
        features = np.random.rand(self.seq_length, self.feat_length)
        label = np.random.randint(2)
        letters = string.ascii_lowercase

        # Generate a unique file name, i.e. one that hasn't been used in this test yet.
        file = ''.join(random.choice(letters) for i in range(10)) + '.mp4'
        while file in self.taken_files:
            file = ''.join(random.choice(letters) for i in range(10)) + '.mp4'
        self.taken_files.add(file)

        return {'feat_seq': features, 'label': label, 'file': file}

    def check_db(self, batch_size, expected_rows, shuffle):
        gen = DatasetGenerator(self.dataset_file_path,
                               self.dataset_name,
                               batch_size,
                               'feat_seq',
                               shuffle=shuffle)
        gen_features = []
        gen_labels = []
        for features, labels in gen.generator(1):
            gen_features.extend(features.tolist())
            gen_labels.extend(labels.tolist())
        self.assertEqual(len(expected_rows), len(gen_labels))

        for gen_label, gen_features in zip(gen_labels, gen_features):
            result = [
                row for row in expected_rows if row['label'] == gen_label
                and np.array_equal(row['feat_seq'], gen_features)
            ]
            self.assertTrue(result)

    def test_full(self):
        expected_rows = self.initialize_expected_rows()
        for i in range(0, self.num_rows):
            row = self.generate_random_row()
            expected_rows[i] = row
            self.dataset_writer.add(row)
        self.dataset_writer.close()

        batch_size = 3
        self.check_db(batch_size, expected_rows, False)

    def test_full_shuffle(self):
        expected_rows = self.initialize_expected_rows()
        for i in range(0, self.num_rows):
            row = self.generate_random_row()
            expected_rows[i] = row
            self.dataset_writer.add(row)
        self.dataset_writer.close()

        batch_size = 3
        self.check_db(batch_size, expected_rows, True)


if __name__ == '__main__':
    unittest.main()



import unittest
from datasetio.datasetwriter import DatasetWriter
import h5py
import os
import numpy as np
import string
import random


class TestDatasetWriter(unittest.TestCase):
    def setUp(self):
        self.feat_length = 10
        self.seq_length = 20
        self.buffer_size = 5
        self.num_rows = 100
        self.dataset_file_path = 'test.hdf'

        self.dtypes=[('feat_seq', 'float', (self.seq_length, self.feat_length)), ('label', 'int'), ('file', h5py.string_dtype())]
        self.dataset_writer = DatasetWriter('test', self.num_rows, self.dtypes, self.dataset_file_path, self.buffer_size)
        self.taken_files = set()

    def tearDown(self):
        os.remove(self.dataset_file_path)

    def initialize_expected_rows(self):
        expected_rows = []
        for i in range(0, self.num_rows):
            zero_features = np.zeros((self.seq_length, self.feat_length))
            row = self.generate_row(zero_features, 0, '')
            expected_rows.append(row)

        return expected_rows

    def generate_row(self, features, label, file):
        return {'feat_seq': features, 'label': label, 'file': file}

    def generate_random_row(self):
        features = np.random.rand(self.seq_length, self.feat_length)
        label = np.random.randint(2)
        letters = string.ascii_lowercase

        # Generate a unique file name, i.e. one that hasn't been used in this test yet.
        file = ''.join(random.choice(letters) for i in range(10)) + '.mp4'
        while file in self.taken_files:
            file = ''.join(random.choice(letters) for i in range(10)) + '.mp4'
        self.taken_files.add(file)

        return {'feat_seq': features, 'label': label, 'file': file}

    def check_equality(self, expected_row, actual_row):
        expected_row_tuple = tuple([expected_row[name] for name in [dtype[0] for dtype in self.dtypes]])
        actual_row_tuple = tuple(actual_row)
        
        for expected_val, actual_val in zip(expected_row_tuple, actual_row_tuple):
            if isinstance(expected_val, np.ndarray):
                if not np.array_equal(expected_val, actual_val):
                    return False
            else:
                if expected_val != actual_val:
                    return False

        return True

    def check_db(self, expected_rows):
        db = h5py.File(self.dataset_file_path, 'r')
        actual_rows = db['test']
        for expected_row, actual_row in zip(expected_rows, actual_rows):
            self.assertTrue(self.check_equality(expected_row, actual_row))

    def test_empty(self):
        expected_rows = self.initialize_expected_rows()
        self.check_db(expected_rows)

    def test_add_one_less_than_buffer_size(self):
        expected_rows = self.initialize_expected_rows()
        for i in range(0, self.buffer_size - 1):
            row = self.generate_random_row()
            expected_rows[i] = row
            self.dataset_writer.add(row)
        self.dataset_writer.close()

        self.check_db(expected_rows)

    def test_add_one_more_than_buffer_size(self):
        expected_rows = self.initialize_expected_rows()
        for i in range(0, self.buffer_size + 1):
            row = self.generate_random_row()
            expected_rows[i] = row
            self.dataset_writer.add(row)
        self.dataset_writer.close()

        self.check_db(expected_rows)

    def test_full(self):
        expected_rows = self.initialize_expected_rows()
        for i in range(0, self.num_rows):
            row = self.generate_random_row()
            expected_rows[i] = row
            self.dataset_writer.add(row)
        self.dataset_writer.close()

        self.check_db(expected_rows)


if __name__ == '__main__':
    unittest.main()



class ResizePreprocessor:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def preprocess(self, imgs):
        return [img.resize((self.width, self.height)) for img in imgs]



from tensorflow.keras.applications.inception_v3 import preprocess_input


class InceptionV3Preprocessor:
    def __init__(self):
        pass

    def preprocess(self, image):
        return preprocess_input(image)



import cv2


class ContrastPreprocessor:
    def __init__(self, alpha, beta):
        self.alpha = alpha
        self.beta = beta

    def preprocess(self, imgs):
        return [
            cv2.convertScaleAbs(img, alpha=self.alpha, beta=self.beta)
            for img in imgs
        ]



from skimage.io import imread
from skimage.transform import resize
import numpy as np
import math
import h5py
from tensorflow.keras.utils import Sequence

# Here, `x_set` is list of path to the images
# and `y_set` are the associated classes.


class DatasetGemeratpr2(Sequence):
    def __init__(self, db_file_path, dataset_name, x_key, y_key, batch_size,
                 preprocessors):
        self.db = h5py.File(db_file_path, "r")
        self.dataset = self.db[dataset_name]
        self.x_key = x_key
        self.y_key = y_key
        self.batch_size = batch_size
        self.preprocessors = preprocessors

    def __len__(self):
        return math.ceil(self.dataset.shape[0] / self.batch_size)

    def __getitem__(self, idx):
        batch_x = self.dataset[idx * self.batch_size:(idx + 1) *
                               self.batch_size][self.x_key]
        batch_y = self.dataset[idx * self.batch_size:(idx + 1) *
                               self.batch_size][self.y_key]

        # TODO: See Trey Hunner's way of making this prettier with a list comprehension
        batch_x_pp = []
        for x in batch_x:
            for pp in preprocessors:
                x = pp.preprocess(x)
            batch_x_pp.append(x)

        return np.array([
            resize(imread(file_name), (200, 200)) for file_name in batch_x
        ]), np.array(batch_y)



# import the necessary packages
from tensorflow.keras.callbacks import BaseLogger
import matplotlib.pyplot as plt
import numpy as np
import json
import os


# Taken from the pyimagesearch Starter Bundle.
class TrainingMonitor(BaseLogger):
    def __init__(self, figPath, jsonPath=None, startAt=0):
        # store the output path for the figure, the path to the JSON
        # serialized file, and the starting epoch
        super(TrainingMonitor, self).__init__()
        self.figPath = figPath
        self.jsonPath = jsonPath
        self.startAt = startAt

    def on_train_begin(self, logs={}):
        # initialize the history dictionary
        self.H = {}

        # if the JSON history path exists, load the training history
        if self.jsonPath is not None:
            if os.path.exists(self.jsonPath):
                self.H = json.loads(open(self.jsonPath).read())

                # check to see if a starting epoch was supplied
                if self.startAt > 0:
                    # loop over the entries in the history log and
                    # trim any entries that are past the starting
                    # epoch
                    for k in self.H.keys():
                        self.H[k] = self.H[k][:self.startAt]

    def on_epoch_end(self, epoch, logs={}):
        # loop over the logs and update the loss, accuracy, etc.
        # for the entire training process
        for (k, v) in logs.items():
            l = self.H.get(k, [])
            l.append(float(v))
            self.H[k] = l

        # check to see if the training history should be serialized
        # to file
        if self.jsonPath is not None:
            f = open(self.jsonPath, "w")
            f.write(json.dumps(self.H))
            f.close()

        # ensure at least two epochs have passed before plotting
        # (epoch starts at zero)
        if len(self.H["loss"]) > 1:
            # plot the training loss and accuracy
            N = np.arange(0, len(self.H["loss"]))
            plt.style.use("ggplot")
            plt.figure()
            plt.plot(N, self.H["loss"], label="train_loss")
            plt.plot(N, self.H["val_loss"], label="val_loss")
            plt.plot(N, self.H["accuracy"], label="train_acc")
            plt.plot(N, self.H["val_accuracy"], label="val_acc")
            plt.title("Training Loss and Accuracy [Epoch {}]".format(
                len(self.H["loss"])))
            plt.xlabel("Epoch #")
            plt.ylabel("Loss/Accuracy")
            plt.legend()

            # save the figure
            plt.savefig(self.figPath)
            plt.close()


import numpy as np
import h5py


# Modified slightly from pyimagesearch's hdf5datasetgenerator.py
class DatasetGenerator:
    def __init__(self,
                 db_file_path,
                 dataset_name,
                 batch_size,
                 feat_key,
                 preprocessors=None,
                 aug=None,
                 classes=2,
                 limit=None,
                 shuffle=False):
        self.batch_size = batch_size
        self.feat_key = feat_key
        self.preprocessors = preprocessors
        self.aug = aug
        self.classes = classes
        self.shuffle = shuffle

        # open the HDF5 database for reading and determine the total
        # number of entries in the database
        self.db = h5py.File(db_file_path, "r")
        self.dataset = self.db[dataset_name]

        if limit:
            self.num_features = limit
        else:
            self.num_features = self.dataset.shape[0]

        self.idxs = np.arange(0, self.num_features, self.batch_size)
        if self.shuffle:
            np.random.shuffle(self.idxs)

    def generator(self, passes=np.inf):
        # initialize the epoch count
        epochs = 0

        # keep looping infinitely -- the model will stop once we have
        # reach the desired number of epochs
        while epochs < passes:
            # loop over the HDF5 dataset
            for idx in self.idxs:
                # extract the features and labels from the HDF dataset
                features = self.dataset[idx:idx +
                                        self.batch_size][self.feat_key]
                labels = self.dataset[idx:idx + self.batch_size]['label']

                # check to see if our preprocessors are not None
                if self.preprocessors is not None:
                    # initialize the list of processed features
                    proc_features = []

                    # loop over the features
                    for feature in features:
                        # loop over the preprocessors and apply each
                        # to the feature
                        for p in self.preprocessors:
                            feature = p.preprocess(feature)

                        # update the list of processed features
                        proc_features.append(feature)

                    # update the features array to be the processed
                    # features
                    features = np.array(proc_features)

            # if the data augmenator exists, apply it
                if self.aug is not None:
                    (features, labels) = next(
                        self.aug.flow(features,
                                      labels,
                                      batch_size=self.batch_size))

            # yield a tuple of features and labels
                yield (features, labels)
            # increment the total number of epochs
            epochs += 1
            if self.shuffle:
                np.random.shuffle(self.idxs)

    def close(self):
        # close the database
        self.db.close()


import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='rocheml',
    version='1.3',
    author='Hayden Roche',
    author_email='haydenroche5@gmail.com',
    description='A package of useful machine learning classes and functions.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.5',
)


import cv2
from facenet_pytorch import MTCNN
from PIL import Image

def boost_contrast(imgs, alpha):
    return [cv2.convertScaleAbs(img , alpha=alpha, beta=0.0) for img in imgs]

# From: https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/.
def bb_iou (boxA, boxB):
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(boxA[0], boxB[0])
    yA = max(boxA[1], boxB[1])
    xB = min(boxA[2], boxB[2])
    yB = min(boxA[3], boxB[3])
    # compute the area of intersection rectangle
    interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)
    # compute the area of both the prediction and ground-truth
    # rectangles
    boxAArea = (boxA[2] - boxA[0] + 1) * (boxA[3] - boxA[1] + 1)
    boxBArea = (boxB[2] - boxB[0] + 1) * (boxB[3] - boxB[1] + 1)
    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = interArea / float(boxAArea + boxBArea - interArea)
    # return the intersection over union value
    return iou

class FaceDetector:
    def __init__(self, min_face_size, margin, prob_threshold):
        self.detector = MTCNN(keep_all=True, post_process=False, device='cuda:0', select_largest=False, min_face_size=min_face_size)
        self.margin = margin
        self.prob_threshold = prob_threshold
        
    def crop_face(self, img, box):
        max_x = img.shape[1] - 1
        max_y = img.shape[0] - 1
        x1, y1, x2, y2 = [int(round(c)) for c in box]
        x1 = max([x1 - self.margin, 0])
        y1 = max([y1 - self.margin, 0])
        x2 = min([x2 + self.margin, max_x])
        y2 = min([y2 + self.margin, max_y])
        face = img[y1:y2, x1:x2]
        new_box = [[x1, y1], [x2, y2]]
        
        return face, new_box

    def filter_faces(self, frame_data):
        faces = {}

        for frame_datum in frame_data:
            for face_data in frame_datum['faces']:
                face_id = face_data['id']
                
                if face_id in faces:
                    faces[face_id].append(face_data['prob'])
                else:
                    faces[face_id] = [face_data['prob']]

        num_frames = len(frame_data)
        face_ids_to_del = set()
        avg_probs = {}

        for face_id, probs in faces.items():
            if len(probs) < (num_frames / 2):
                print(f'FaceDetector::filter_faces: Face with id {face_id} failed to appear in >= {num_frames / 2} frames.')
                face_ids_to_del.add(face_id)
            avg_probs[face_id] = sum(probs) / len(probs)

        faces_remaining = len(faces) - len(face_ids_to_del)

        if faces_remaining > 2:
            print('FaceDetector::filter_faces: More than 2 faces. Only keeping the two with the highest avg prob.')
            avg_probs_sorted = sorted(avg_probs.items(), key=lambda x: x[1], reverse=True)

            for face_id, avg_prob in avg_probs_sorted[2:]:
                face_ids_to_del.add(face_id)

        filtered_frame_data = []
        for frame_datum in frame_data:
            faces_filtered = []

            for face_data in frame_datum['faces']:
                if face_data['id'] in face_ids_to_del:
                    continue
                else:
                    faces_filtered.append(face_data)

            frame_datum['faces'] = faces_filtered
            filtered_frame_data.append(frame_datum)

        return filtered_frame_data
    
    def detect(self, video, num_frames, filt=True):
        vc = cv2.VideoCapture(video)        
        imgs = []
        
        for i in range(num_frames):
            success, img = vc.read()
            if success:
                imgs.append(img)
            else:
                print('FaceDetector::detect: cv2::VideoCapture::read call failed.')
                return []
        
        imgs_pil = [Image.fromarray(i) for i in imgs]
        video_boxes, video_probs, video_landmarks = self.detector.detect(imgs_pil, landmarks=True)
        past_faces = []
        frame_data = []
        face_id = 0
        
        def search_past_faces(box, iou_threshold):
            i = 0
            for past_face in past_faces:
                iou = bb_iou(past_face['box'], face_box)
                if iou > iou_threshold:
                    return (True, i)
                else:
                    i += 1
            return (False, i)
            
        
        for frame, (frame_boxes, frame_probs, frame_landmarks) in enumerate(zip(video_boxes, video_probs, video_landmarks)):
            contrast_boosted = False
            img = imgs[frame]

            if frame_boxes is None:
#                 print(f'FaceDetector::detect: No faces in frame {frame}. Boosting contrast.')
                img = boost_contrast([img], 3.0)[0]
                img_pil = Image.fromarray(img)
                frame_boxes, frame_probs, frame_landmarks = self.detector.detect(img_pil, landmarks=True)
                contrast_boosted = True
                
                if frame_boxes is None:
#                     print(f'FaceDetector::detect: No faces after contrast boost. Proceeding to next frame.')
                    frame_data.append({'frame': frame, 'faces': [], 'contrast_boosted': contrast_boosted})
                    continue

            face_data = []
            
            for face_box, face_prob, face_landmarks in zip(frame_boxes, frame_probs, frame_landmarks):
                if face_prob < self.prob_threshold:
                    continue
               
                face_cropped, new_box = self.crop_face(img, face_box)
                found, idx = search_past_faces(face_box, 0.5)
                
                if found:
                    past_faces[idx]['box'] = face_box
                    face_data.append({'id': past_faces[idx]['id'], 'box': new_box, 'prob': face_prob, 'landmarks': face_landmarks, 'img': face_cropped})
                else:
                    past_faces.append({'box': face_box , 'id': face_id})
                    face_data.append({'id': face_id, 'box': new_box, 'prob': face_prob, 'landmarks': face_landmarks, 'img': face_cropped})
                    face_id += 1

            frame_data.append({'frame': frame, 'faces': face_data, 'contrast_boosted': contrast_boosted})

        return frame_data if not filt else self.filter_faces(frame_data) 
                